#!/usr/bin/env python
import numpy as np
from collections.abc import Iterable
from collections import defaultdict
import matplotlib.pyplot as plt
from ase.dft.kpoints import bandpath
from TB2J.kpoints import monkhorst_pack
from .hamiltonian_terms import (
    ZeemanTerm,
    UniaxialMCATerm,
    ExchangeTerm,
    DMITerm,
    BilinearTerm,
)
from .constants import mu_B, gyromagnetic_ratio
from .supercell import SupercellMaker
from .spin_xml import SpinXmlParser, SpinXmlWriter
from .plot import group_band_path
from ase.cell import Cell
from .qsolver import QSolver
import json
import os


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


plot_script = """\
#!/usr/bin/env python
\"\"\"
Parse the json file generated by the magnon band calculation. 
And plot the magnon band structure from the json file. 
You can modify this file to plot the magnon band structure, all plot 
multiple magnon band structures in one figure.
\"\"\"
import numpy as np
import matplotlib.pyplot as plt
import json

def load_magnon_json(filename):
    \"\"\"
    Load the magnon band structure from the json file.
    args:
        filename: the json file generated by the magnon band calculation.
    return:
        kptlist: all the kpoints
        xlist: the x coordinates of the kpoints in the plot
        knames: names of the high symmetry points
        Xs: The x coordinates of the high symmetry points in the plot
        allevals: eigenvalues, in meV
            Note that the eigenvalues are splitted into groups of kpoints.
            the first index is the index of the group of kpoints
            the second index is the index of the kpoint in the group 
            the third index is the index of the band
    \"\"\"
    with open(filename) as f:
        data = json.load(f)
    kptlist = data["kptlist"]
    xlist = data["xlist"]
    knames = data["knames"]
    Xs = data["X_for_highsym_kpoints"]
    allevals = [np.array(x) for x in data["evals"]]
    return kptlist, xlist, knames, Xs, allevals

def plot_magnon_json(filename, ax, color="k", show=True):
    \"\"\"
    Plot the magnon band structure from the json file.
    args:
        filename: the json file generated by the magnon band calculation.
        ax: the matplotlib ax to plot the band structure.
            If None, a new figure will be created.
        color: the color of the band structure.
        show: whether to show the plot. 
    \"\"\"
    if ax is None:
        fig, ax = plt.subplots()

    kptlist, xlist, knames, Xs, allevals = load_magnon_json(filename)
    for i, (kpts, xs) in enumerate(zip(kptlist, xlist)):
        evals=allevals[i]
        # Plot band structure
        nbands = evals.shape[1]
        emin = np.min(evals[:, 0])
        for i in range(nbands):
            ax.plot(xs, (evals[:, i]) / 1.6e-22, color=color)

    ax.set_ylabel("Energy (meV)")
    ax.set_xlim(xlist[0][0], xlist[-1][-1])
    ax.set_xticks(Xs)
    knames = [x if x != "G" else "$\Gamma$" for x in knames]
    ax.set_xticklabels(knames)
    for x in Xs:
        ax.axvline(x, linewidth=0.6, color="gray")
    if show:
        plt.show()


if __name__ == "__main__":
    plot_magnon_json("magnon_band.json", ax=None, color="k", show=True)

"""


class SpinHamiltonian(object):
    def __init__(
        self,
        cell=None,
        pos=None,
        xcart=None,
        spinat=None,
        zion=None,
        Rlist=None,
        iprim=None,
    ):
        if xcart is None:
            self.xcart = np.dot(pos, cell)
            self.pos = pos
        elif pos is None:
            self.pos = np.dot(xcart, np.linalg.inv(cell))
            self.xcart = xcart
        self.cell = cell
        self.xcart = pos
        self.spinat = spinat
        self.zion = zion

        self._spin = np.array(spinat) * mu_B
        self._map_to_magnetic_only()
        # spin interaction parameters

        # Homogeneous H eff_field
        self.has_external_hfield = False
        self.zeeman_H = np.zeros(3, dtype="float64")

        # uniaxial single ion anisotropy
        self.has_uniaxial_anisotropy = False

        # cubic Anistropy (Not implemented yet)
        self.has_cubic_anistropy = False

        # Exchange interaction
        self.has_exchange = False
        self.exchange_Jdict = None

        # DM interaction
        self.has_dmi = False
        self.dmi_ddict = None

        # Dipole dipole interaction
        self.has_dipdip = False

        # bilinear term
        self.has_bilinear = False
        self.bilinear_dict = None

        # calculation parameters

        # calculation results

        # hamiltonian list
        self.hamiltonians = {}

        self._total_hessian_ijR = None

        self.gilbert_damping = np.ones(self.nspin) * 1.0
        self.gyro_ratio = np.ones(self.nspin) * gyromagnetic_ratio

        # internal variables:
        self._langevin_tmp = None

        self.sublattice = list(range(self.nspin))

        self.index_spin = None
        self.lattice = None

        self.iprim = iprim
        self.Rlist = Rlist

    def set_lattice(self, atoms, index_spin):
        self.lattice = atoms
        self.index_spin = index_spin

    @property
    def spin_positions(self):
        return self.pos

    def _map_to_magnetic_only(self):
        """
        select the sites with spin and re-index.
        """
        ms = np.linalg.norm(self._spin, axis=1)
        self.magsites = np.where(ms > -0.00001 * mu_B)[0]
        self.ms = ms[self.magsites]
        S = self._spin[self.magsites]
        self.nspin = len(S)
        # self.S = np.array(
        #    [S[i] / self.ms[i] for i in range(self.nspin)],
        #    dtype='float64')
        self.s = S / self.ms[:, None]

    @property
    def spin(self):
        self._spin[self.magsites] = np.multiply(self.ms[:, np.newaxis], self.s) / mu_B
        return self._spin

    @spin.setter
    def spin(self, spin):
        self._spin = np.array(spin) * mu_B
        self._map_to_magnetic_only()

    def normalize_S(self):
        """
        normalize so the norm of self.S[i] is 1
        """
        snorm = np.linalg.norm(self.s, axis=1)
        self.s /= np.expand_dims(snorm, axis=1)

    def set(
        self,
        gilbert_damping=None,
        gyro_ratio=None,
        has_exchange=None,
        has_dmi=None,
        has_bilinear=None,
        has_external_hfield=None,
        has_uniaxial_anisotropy=None,
    ):
        """
        set parameters for simulation:
        args:
        ====================
        timestep: in ps
        temperature: in K, default 0K.
        damping factor: Gilbert damping factor, default: 0.01
        gyromagnetic_ratio: default 1.0.
        """
        if gilbert_damping is not None:
            self.gilbert_damping = np.array(gilbert_damping)
        if gyro_ratio is not None:
            self.gyro_ratio = np.array(gyro_ratio)
        self.has_exchange = self.has_exchange and bool(has_exchange)
        self.has_dmi = self.has_dmi and bool(has_dmi)
        self.has_bilinear = self.has_bilinear and bool(has_bilinear)
        self.has_external_hfield = self.has_external_hfield and bool(
            has_external_hfield
        )
        self.has_uniaxial_anisotropy = self.has_uniaxial_anisotropy and bool(
            has_uniaxial_anisotropy
        )

    def set_exchange_ijR(self, exchange_Jdict):
        """
        J: [(i,j, R, J_{ijR})] J_{ijR} is a scalar
        """
        self.has_exchange = True
        self.exchange_Jdict = exchange_Jdict
        exchange = ExchangeTerm(self.exchange_Jdict, ms=self.ms)
        self.hamiltonians["exchange"] = exchange

    def set_dmi_ijR(self, dmi_ddict):
        """
        D: [(i,j, R, D_{ijR})], D_{ijR} is a vector
        """
        self.has_dmi = True
        self.dmi_ddict = dmi_ddict
        DMI = DMITerm(self.dmi_ddict, self.ms)
        self.hamiltonians["DMI"] = DMI

    def set_bilinear_ijR(self, bilinear_dict):
        self.has_bilinear = True
        self.bilinear_dict = bilinear_dict
        Bi = BilinearTerm(self.bilinear_dict, self.ms)
        self.hamiltonians["Bilinear"] = Bi

    def set_dipdip(self):
        """
        add the dipole dipole interaction term.
        """

    def set_external_hfield(self, H):
        """
        add external magnetic field. If H is a vector , it is homogenoues.
            Otherwise H should be given as a nspin*3 matrix.
        """
        self.has_external_hfield = True
        if isinstance(H, Iterable):
            self.H_ext = np.asarray(H)
        else:
            self.H_ext = np.ones([self.nspin, 3]) * H
        zeeman = ZeemanTerm(H=self.H_ext, ms=self.ms)
        self.hamiltonians["zeeman"] = zeeman

    def set_uniaxial_mca(self, k1, k1dir):
        """
        Add homogenoues uniaxial anisotropy
        """
        self.has_uniaxial_anisotropy = True

        self.k1 = k1
        self.k1dir = k1dir
        umcaterm = UniaxialMCATerm(k1, k1dir, ms=self.ms)
        self.hamiltonians["UMCA"] = umcaterm

    def add_Hamiltonian_term(self, Hamiltonian_term, name=None):
        """
        add Hamiltonian term which is not pre_defined.
        """
        if name in self.hamiltonians:
            raise ValueError(
                "Hamiltonian name %s already defined. The defined names are %s"
                % (name, self.hamiltonians.keys())
            )
        else:
            self.hamiltonians[name] = Hamiltonian_term

    # @profile
    def get_effective_field(self, S):
        """
        calculate the effective field Heff=-1/ms * \partial H / \partial S
        Langevin term not included.
        """
        Heff = 0.0
        for ham in self.hamiltonians.values():
            Heff += ham.eff_field(S)
        return Heff

    def make_supercell(self, sc_matrix=None, supercell_maker=None):
        if supercell_maker is None:
            smaker = SupercellMaker(sc_matrix)
        else:
            smaker = supercell_maker

        sc_cell = smaker.sc_cell(np.array(self.cell))
        sc_pos = np.array(smaker.sc_pos(np.array(self.pos)))
        sc_zion = smaker.sc_trans_invariant(np.array(self.zion))

        if self.index_spin is not None:
            smaker.sc_trans_invariant(self.index_spin)

        sc_Rlist = np.repeat(smaker.R_sc, self.nspin, axis=0)
        sc_iprim = smaker.sc_trans_invariant(list(range(self.nspin)))

        sc_spinat = np.array(smaker.sc_trans_invariant(self.spinat))

        sc_ham = SpinHamiltonian(
            cell=sc_cell,
            pos=sc_pos,
            spinat=sc_spinat,
            zion=sc_zion,
            Rlist=sc_Rlist,
            iprim=sc_iprim,
        )

        sc_gyro_ratio = np.array(smaker.sc_trans_invariant(self.gyro_ratio))
        sc_ham.gyro_ratio = sc_gyro_ratio

        sc_gilbert_damping = np.array(smaker.sc_trans_invariant(self.gilbert_damping))
        sc_ham.gilbert_damping = sc_gilbert_damping

        if self.has_external_hfield:
            sc_Hext = smaker.sc_trans_invariant(self.H_ext)
            sc_ham.set_external_hfield(sc_Hext)

        if self.has_uniaxial_anisotropy:
            sc_k1 = smaker.sc_trans_invariant(self.k1)
            sc_k1dir = smaker.sc_trans_invariant(self.k1dir)
            sc_ham.set_uniaxial_mca(sc_k1, np.array(sc_k1dir))

        if self.has_exchange:
            sc_Jdict = smaker.sc_ijR(self.exchange_Jdict, n_basis=len(self.pos))
            sc_ham.set_exchange_ijR(exchange_Jdict=sc_Jdict)

        if self.has_dmi:
            sc_dmi_ddict = smaker.sc_ijR(self.dmi_ddict, n_basis=len(self.pos))
            sc_ham.set_dmi_ijR(sc_dmi_ddict)

        if self.has_bilinear:
            sc_bilinear_dict = smaker.sc_ijR(self.bilinear_dict, n_basis=len(self.pos))
            sc_ham.set_bilinear_ijR(sc_bilinear_dict)

        return sc_ham

    def calc_total_HijR(self):
        self._total_hessian_ijR = defaultdict(lambda: np.zeros((3, 3), dtype=float))
        for tname, term in self.hamiltonians.items():
            if term.is_twobody_term():
                for key, val in term.hessian_ijR().items():
                    self._total_hessian_ijR[key] += val
        return self._total_hessian_ijR

    def get_total_hessian_ijR(self):
        if self._total_hessian_ijR is None:
            self.calc_total_HijR()
        return self._total_hessian_ijR

    def solve_k(self, kpts, Jq=False):
        """
        Get the eigenvalues and eigenvectors for the kpoints
        """
        qsolver = QSolver(hamiltonian=self)
        evals, evecs = qsolver.solve_all(kpts, eigen_vectors=True, Jq=Jq)
        return evals, evecs

    def find_ground_state_from_kmesh(self, kmesh, myfile):
        kpts = monkhorst_pack(kmesh)
        evals, evecs = self.solve_k(kpts, Jq=True)
        # write_magnon_info(self, kpts, evals, evecs, myfile)

    def plot_magnon_band(
        self,
        kvectors=np.array(
            [[0, 0, 0], [0.5, 0, 0], [0.5, 0.5, 0], [0, 0, 0], [0.5, 0.5, 0.5]]
        ),
        knames=["$\Gamma$", "X", "M", "$\Gamma$", "R"],
        supercell_matrix=None,
        npoints=50,
        color="red",
        kpath_fname=None,
        Jq=False,
        ax=None,
        output_fname="magnon_band.json",
    ):
        if ax is None:
            fig, ax = plt.subplots()
        kptlist = kvectors
        if knames is None and kvectors is None:
            # fully automatic k-path
            bp = Cell(self.cell).bandpath(npoints=npoints)
            spk = bp.special_points
            xlist, kptlist, Xs, knames = group_band_path(bp)
        elif knames is not None and kvectors is None:
            # user specified kpath by name
            bp = Cell(self.cell).bandpath(knames, npoints=npoints)
            spk = bp.special_points
            kpts = bp.kpts
            xlist, kptlist, Xs, knames = group_band_path(bp)
        else:
            # user spcified kpath and kvector.
            kpts, x, Xs = bandpath(kvectors, self.cell, npoints)
            spk = dict(zip(knames, kvectors))
            xlist = [x]
            kptlist = [kpts]

        if supercell_matrix is not None:
            kptlist = [np.dot(k, supercell_matrix) for k in kptlist]
        print("High symmetry k-points:")
        for name, k in spk.items():
            if name == "G":
                name = "Gamma"
            print(f"{name}: {k}")

        allevals = []
        for kpts, xs in zip(kptlist, xlist):
            evals, evecs = self.solve_k(kpts, Jq=Jq)
            allevals.append(evals)
            # Plot band structure
            nbands = evals.shape[1]
            emin = np.min(evals[:, 0])
            for i in range(nbands):
                ax.plot(xs, (evals[:, i]) / 1.6e-22, color=color)

        ax.set_ylabel("Energy (meV)")
        ax.set_xlim(xlist[0][0], xlist[-1][-1])
        ax.set_xticks(Xs)
        knames = [x if x != "G" else "$\Gamma$" for x in knames]
        ax.set_xticklabels(knames)
        for x in Xs:
            ax.axvline(x, linewidth=0.6, color="gray")

        self.magnon_info = {
            "kptlist": kptlist,
            "xlist": xlist,
            "knames": knames,
            "X_for_highsym_kpoints": Xs,
            "knames": knames,
            "nbands": nbands,
            "nkpts": len(kptlist),
            "evals": allevals,
        }
        if output_fname is not None:
            self.save_magnon_json(output_fname)
        return ax

    def save_magnon_json(self, fname):
        with open(fname, "w") as myfile:
            json.dump(self.magnon_info, myfile, cls=NumpyEncoder, indent=4)
        print(
            f"""Saved magnon info to {fname}.
Use the following script to plot: 
python plot_magnon_from_json_file.py
Or modify the script to plot in the way you want. 
"""
        )
        with open("plot_magnon_from_json_file.py", "w") as myfile:
            myfile.write(plot_script)

    def write_xml(self, fname):
        writer = SpinXmlWriter()
        writer._write(self, fname)


def read_spin_ham_from_file(fname):
    parser = SpinXmlParser(fname)
    ham = SpinHamiltonian(
        cell=parser.cell,
        xcart=parser.spin_positions,
        spinat=parser.spin_spinat,
        zion=parser.spin_zions,
    )
    ham.set(
        gilbert_damping=parser.spin_damping_factors, gyro_ratio=parser.spin_gyro_ratios
    )
    if parser.has_exchange:
        ham.set_exchange_ijR(parser.exchange(isotropic=True))
    if parser.has_dmi:
        ham.set_dmi_ijR(parser.dmi)
    if parser.has_bilinear:
        ham.set_bilinear_ijR(parser.bilinear)
    return ham



from collections import Counter

import numpy as np
import spglib


def monkhorst_pack(size, gamma_center=False):
    """Construct a uniform sampling of k-space of given size.
    Modified from ase.dft.kpoints with gamma_center option added"""
    if np.less_equal(size, 0).any():
        raise ValueError("Illegal size: %s" % list(size))
    kpts = np.indices(size).transpose((1, 2, 3, 0)).reshape((-1, 3))
    asize = np.array(size)
    shift = 0.5 * ((asize + 1) % 2) / asize
    mkpts = (kpts + 0.5) / size - 0.5
    if gamma_center:
        mkpts += shift
    return mkpts


def get_ir_kpts(atoms, mesh):
    """
    Gamma-centered IR kpoints. mesh : [nk1,nk2,nk3].
    """
    lattice = atoms.get_cell()
    positions = atoms.get_scaled_positions()
    numbers = atoms.get_atomic_numbers()

    cell = (lattice, positions, numbers)
    mapping, grid = spglib.get_ir_reciprocal_mesh(mesh, cell, is_shift=[0, 0, 0])
    return grid[np.unique(mapping)] / np.array(mesh, dtype=float)


def ir_kpts(
    atoms, mp_grid, is_shift=[0, 0, 0], verbose=True, ir=True, is_time_reversal=False
):
    """
    generate kpoints for structure
    Parameters:
    ------------------
    atoms: ase.Atoms
      structure
    mp_grid: [nk1,nk2,nk3]
    is_shift: shift of k points. default is Gamma centered.
    ir: bool
    Irreducible or not.
    """
    cell = (atoms.get_cell(), atoms.get_scaled_positions(), atoms.get_atomic_numbers())
    # print(spglib.get_spacegroup(cell, symprec=1e-5))
    mesh = mp_grid
    # Gamma centre mesh
    mapping, grid = spglib.get_ir_reciprocal_mesh(
        mesh, cell, is_shift=is_shift, is_time_reversal=is_time_reversal, symprec=1e-4
    )
    if not ir:
        return (np.array(grid).astype(float) + np.asarray(is_shift) / 2.0) / mesh, [
            1.0 / len(mapping)
        ] * len(mapping)
    # All k-points and mapping to ir-grid points
    # for i, (ir_gp_id, gp) in enumerate(zip(mapping, grid)):
    #    print("%3d ->%3d %s" % (i, ir_gp_id, gp.astype(float) / mesh))
    cnt = Counter(mapping)
    ids = list(cnt.keys())
    weight = list(cnt.values())
    weight = np.array(weight) * 1.0 / sum(weight)
    ird_kpts = [
        (grid[id].astype(float) + np.asarray(is_shift) / 2.0) / mesh for id in ids
    ]

    # Irreducible k-points
    # print("Number of ir-kpoints: %d" % len(np.unique(mapping)))
    # print(grid[np.unique(mapping)] / np.array(mesh, dtype=float))

    new_ir_kpts = []
    new_weight = []
    for ik, k in enumerate(ird_kpts):
        # add k and -k to ird_kpts if -k is not in ird_kpts
        if not any([np.allclose(-1.0 * k, kpt) for kpt in new_ir_kpts]):
            new_ir_kpts.append(k)
            new_ir_kpts.append(-1.0 * k)
            new_weight.append(weight[ik] / 2)
            new_weight.append(weight[ik] / 2)
        else:
            new_ir_kpts.append(k)
            new_weight.append(weight[ik])
    # return ird_kpts, weight
    return np.array(new_ir_kpts), np.array(new_weight)


def test_ir_kpts():
    from ase.build import bulk

    atoms = bulk("Si")
    mesh = [14, 14, 14]
    kpts = get_ir_kpts(atoms, mesh)
    print(kpts)
    print(len(kpts))


if __name__ == "__main__":
    test_ir_kpts()



"""
This module provides functionality for handling magnetic exchange interactions and computing magnon band structures.

It includes classes and functions for:
- Reading and writing exchange interaction data
- Computing exchange tensors and magnon energies
- Plotting magnon band structures
- Converting between different magnetic structure representations
"""

import numpy as np
from scipy.spatial.transform import Rotation

from .magnon_math import generate_grid, get_rotation_arrays, round_to_precision, uz
from .plot import BandsPlot
from .structure import BaseMagneticStructure, get_attribute_array

__all__ = [
    "ExchangeIO",
    "plot_tb2j_magnon_bands",
]


def branched_keys(tb2j_keys, npairs):
    """
    Organize TB2J keys into branches based on magnetic site pairs.

    Parameters
    ----------
    tb2j_keys : list
        List of TB2J dictionary keys containing interaction information
    npairs : int
        Number of magnetic site pairs

    Returns
    -------
    list
        List of branched keys organized by magnetic site pairs
    """
    msites = int((2 * npairs) ** 0.5)
    branch_size = len(tb2j_keys) // msites**2
    new_keys = sorted(tb2j_keys, key=lambda x: -x[1] + x[2])[
        (npairs - msites) * branch_size :
    ]
    new_keys.sort(key=lambda x: x[1:])
    bkeys = [
        new_keys[i : i + branch_size] for i in range(0, len(new_keys), branch_size)
    ]

    return [sorted(branch, key=lambda x: np.linalg.norm(x[0])) for branch in bkeys]


def correct_content(content, quadratic=False):
    """
    Ensure content dictionary has all required entries with proper initialization.

    Parameters
    ----------
    content : dict
        Dictionary containing exchange interaction data
    quadratic : bool, optional
        Whether to include biquadratic interactions, by default False
    """
    n = max(content["index_spin"]) + 1
    data_shape = {"exchange_Jdict": ()}

    if not content["colinear"]:
        data_shape |= {"Jani_dict": (3, 3), "dmi_ddict": (3,)}
    if quadratic:
        data_shape["biquadratic_Jdict"] = (2,)

    for label, shape in data_shape.items():
        content[label] |= {((0, 0, 0), i, i): np.zeros(shape) for i in range(n)}


def Hermitize(array):
    """
    Convert an array into its Hermitian form by constructing a Hermitian matrix.

    A Hermitian matrix H has the property that H = H†, where H† is the conjugate transpose.
    This means H[i,j] = conj(H[j,i]) for all indices i,j. The function takes an input array
    representing the upper triangular part of the matrix and constructs the full Hermitian
    matrix by:
    1. Placing the input values in the upper triangular part
    2. Computing the conjugate transpose of these values for the lower triangular part

    This is commonly used in quantum mechanics and magnetic systems where Hamiltonians
    must be Hermitian to ensure real eigenvalues.

    Parameters
    ----------
    array : numpy.ndarray
        Input array containing the upper triangular elements of the matrix.
        Shape should be (n*(n+1)/2, ...) where n is the dimension of
        the resulting square matrix.

    Returns
    -------
    numpy.ndarray
        Full Hermitian matrix with shape (n, n, ...), where n is computed
        from the input array size. The result satisfies result[i,j] = conj(result[j,i])
        for all indices i,j.

    Example
    -------
    >>> arr = np.array([1+0j, 2+1j, 3+0j])  # Upper triangular elements for 2x2 matrix
    >>> Hermitize(arr)
    array([[1.+0.j, 2.+1.j],
           [2.-1.j, 3.+0.j]])
    """
    n = int((2 * array.shape[0]) ** 0.5)
    result = np.zeros((n, n) + array.shape[1:], dtype=complex)
    u_indices = np.triu_indices(n)

    # for python>=3.11
    # result[*u_indices] = array
    # result.swapaxes(0, 1)[*u_indices] = array.swapaxes(-1, -2).conj()
    # for python<3.11
    result[u_indices[0], u_indices[1]] = array
    result.swapaxes(0, 1)[u_indices[0], u_indices[1]] = array.swapaxes(-1, -2).conj()

    return result


class ExchangeIO(BaseMagneticStructure):
    """
    Class for handling magnetic exchange interactions and computing magnon properties.

    This class provides functionality for:
    - Managing magnetic structure information
    - Computing exchange tensors
    - Calculating magnon band structures
    - Reading TB2J format files
    - Visualizing magnon bands

    Parameters
    ----------
    atoms : ase.Atoms, optional
        ASE atoms object containing the structure
    cell : array_like, optional
        3x3 matrix defining the unit cell
    elements : list, optional
        List of chemical symbols for atoms
    positions : array_like, optional
        Atomic positions
    magmoms : array_like, optional
        Magnetic moments for each atom
    pbc : tuple, optional
        Periodic boundary conditions, default (True, True, True)
    magnetic_elements : list, optional
        List of magnetic elements in the structure
    kmesh : list, optional
        k-point mesh dimensions, default [1, 1, 1]
    collinear : bool, optional
        Whether the magnetic structure is collinear, default True
    """

    def __init__(
        self,
        atoms=None,
        cell=None,
        elements=None,
        positions=None,
        magmoms=None,
        pbc=(True, True, True),
        magnetic_elements=[],
        kmesh=[1, 1, 1],
        collinear=True,
    ):
        super().__init__(
            atoms=atoms,
            cell=cell,
            elements=elements,
            positions=positions,
            pbc=pbc,
            magmoms=magmoms,
            collinear=collinear,
        )

        self.magnetic_elements = magnetic_elements
        self.kmesh = kmesh

        num_terms = 4 if collinear else 18
        self._exchange_values = np.empty((0, 0, num_terms), dtype=float)

    @property
    def magnetic_elements(self):
        """List of magnetic elements in the structure."""
        return self._magnetic_elements

    @magnetic_elements.setter
    def magnetic_elements(self, value):
        from .structure import validate_symbols

        symbols = validate_symbols(value)
        for symbol in symbols:
            if symbol not in self.elements:
                raise ValueError(f"Symbol '{symbol}' not in 'elements'.")

        self._magnetic_elements = symbols
        self._set_index_pairs()

    @property
    def interacting_pairs(self):
        """List of pairs of interacting magnetic sites."""
        return self._pairs

    def _set_index_pairs(self):
        from itertools import combinations_with_replacement

        magnetic_elements = self.magnetic_elements
        elements = self.elements
        indices = [
            i for i, symbol in enumerate(elements) if symbol in magnetic_elements
        ]
        index_pairs = list(combinations_with_replacement(indices, 2))
        index_spin = np.sort(np.unique(index_pairs))

        self._pairs = index_pairs
        self._index_spin = index_spin

    @property
    def kmesh(self):
        """K-point mesh dimensions for sampling the Brillouin zone."""
        return self._kmesh

    @kmesh.setter
    def kmesh(self, values):
        try:
            the_kmesh = [int(k) for k in values]
        except (ValueError, TypeError):
            raise ValueError("Argument must be an iterable with 'int' elements.")
        if len(the_kmesh) != 3:
            raise ValueError("Argument must be of length 3.")
        if any(k < 1 for k in the_kmesh):
            raise ValueError("Argument must contain only positive numbers.")

        self._kmesh = the_kmesh

    @property
    def vectors(self):
        """Array of interaction vectors between magnetic sites."""
        return self._exchange_values[:, :, :3]

    def set_vectors(self, values=None, cartesian=False):
        """
        Set the interaction vectors between magnetic sites.

        Parameters
        ----------
        values : array_like, optional
            Array of interaction vectors
        cartesian : bool, optional
            Whether the vectors are in Cartesian coordinates, default False
        """
        try:
            pairs = self._pairs
        except AttributeError:
            raise AttributeError("'magnetic_elements' attribute has not been set yet.")
        else:
            n_pairs = len(pairs)

        if values is None:
            i, j = zip(*pairs)
            positions = self.positions
            base_vectors = positions[i, :] - positions[j, :]
            grid = generate_grid(self.kmesh)
            vectors = base_vectors[:, None, :] + grid[None, :, :]
            m_interactions = np.prod(self.kmesh)
        else:
            vectors = get_attribute_array(values, "vectors", dtype=float)
            if vectors.ndim != 3 or vectors.shape[::2] != (n_pairs, 3):
                raise ValueError(
                    f"'vectors' must have the shape (n, m, ), where n={n_pairs} is the number of\n"
                    "pairs of interacting species."
                )
            if cartesian:
                vectors = np.linalg.solve(self.cell.T, vectors.swapaxes(1, -1))
                vectors = vectors.swapaxes(-1, 1)
            m_interactions = vectors.shape[1]

        shape = (
            (n_pairs, m_interactions, 4)
            if self.collinear
            else (n_pairs, m_interactions, 18)
        )
        exchange_values = np.zeros(shape, dtype=float)
        exchange_values[:, :, :3] = vectors
        self._exchange_values = exchange_values

    def _get_neighbor_indices(self, neighbors, tol=1e-4):
        """
        Get indices of neighbor pairs based on distance.

        Parameters
        ----------
        neighbors : list
            List of neighbor shells to consider
        tol : float, optional
            Distance tolerance for neighbor shell assignment, default 1e-4

        Returns
        -------
        tuple
            Indices corresponding to the specified neighbor shells
        """
        distance = np.linalg.norm(self.vectors @ self.cell, axis=-1)
        distance = round_to_precision(distance, tol)
        neighbors_distance = np.unique(np.sort(distance))
        indices = np.where(
            distance[:, :, None] == neighbors_distance[neighbors][None, None, :]
        )

        return indices

    def set_exchange_array(self, name, values, neighbors=None, tol=1e-4):
        """
        Set exchange interaction values for specified neighbors.

        Parameters
        ----------
        name : str
            Type of exchange interaction ('Jiso', 'Biquad', 'DMI', or 'Jani')
        values : array_like
            Exchange interaction values
        neighbors : list, optional
            List of neighbor shells to assign values to
        tol : float, optional
            Distance tolerance for neighbor shell assignment, default 1e-4
        """
        if self.vectors.size == 0:
            raise AttributeError("The intraction vectors must be set first.")

        array = get_attribute_array(values, name, dtype=float)

        if neighbors is not None:
            if len(array) != len(neighbors):
                raise ValueError(
                    "The number of neighbors and exchange values does not coincide."
                )
            *array_indices, value_indices = self._get_neighbor_indices(
                list(neighbors), tol=tol
            )
        else:
            if self._exchange_values.shape[:2] != array.shape[:2]:
                raise ValueError(
                    f"The shape of the array is incompatible with '{self.exchange_values.shape}'"
                )
            array_indices, value_indices = (
                [slice(None), slice(None)],
                (slice(None), slice(None)),
            )

        if name == "Jiso":
            self._exchange_values[*array_indices, 3] = array[value_indices]
        elif name == "Biquad":
            self._exchange_values[*array_indices, 4:6] = array[value_indices]
        elif name == "DMI":
            self._exchange_values[*array_indices, 6:9] = array[value_indices]
        elif name == "Jani":
            self._exchange_values[*array_indices, 9:] = array[value_indices].reshape(
                array.shape[:2] + (9,)
            )
        else:
            raise ValueError(f"Unrecognized exchange array name: '{name}'.")

    @property
    def Jiso(self):
        return self._exchange_values[:, :, 3]

    @property
    def Biquad(self):
        return self._exchange_values[:, :, 4:6]

    @property
    def DMI(self):
        return self._exchange_values[:, :, 6:9]

    @property
    def Jani(self):
        matrix_shape = self._exchange_values.shape[:2] + (3, 3)
        return self._exchange_values[:, :, 9:].reshape(matrix_shape)

    def exchange_tensor(self, anisotropic=True):
        """
        Compute the exchange interaction tensor.

        Parameters
        ----------
        anisotropic : bool, optional
            Whether to include anisotropic interactions, default True

        Returns
        -------
        numpy.ndarray
            Exchange interaction tensor
        """
        shape = self._exchange_values.shape[:2] + (3, 3)
        tensor = np.zeros(shape, dtype=float)

        if anisotropic and not self.collinear:
            # anisotropic exchange tensor
            tensor += self._exchange_values[:, :, 9:].reshape(shape)
            # DMI
            pos_indices = ([1, 2, 0], [2, 0, 1])
            neg_indices = ([2, 0, 1], [1, 2, 0])
            tensor[:, :, *pos_indices] += self._exchange_values[:, :, 6:9]
            tensor[:, :, *neg_indices] -= self._exchange_values[:, :, 6:9]
        # isotropic exchange
        diag_indices = ([0, 1, 2], [0, 1, 2])
        tensor[:, :, *diag_indices] += self._exchange_values[:, :, 3, None]

        return tensor

    def Jq(self, kpoints, anisotropic=True):
        """
        Compute the exchange interactions in reciprocal space.

        Parameters
        ----------
        kpoints : array_like
            k-points at which to evaluate the exchange interactions
        anisotropic : bool, optional
            Whether to include anisotropic interactions, default True

        Returns
        -------
        numpy.ndarray
            Exchange interactions in reciprocal space
        """
        vectors = self._exchange_values[:, :, :3].copy()
        tensor = self.exchange_tensor(anisotropic=anisotropic)

        if self._Q is not None:
            phi = 2 * np.pi * vectors.round(3).astype(int) @ self._Q
            rv = np.einsum("ij,k->ijk", phi, self._n)
            R = (
                Rotation.from_rotvec(rv.reshape(-1, 3))
                .as_matrix()
                .reshape(vectors.shape[:2] + (3, 3))
            )
            np.einsum("nmij,nmjk->nmik", tensor, R, out=tensor)

        exp_summand = np.exp(2j * np.pi * vectors @ kpoints.T)
        Jexp = exp_summand[:, :, :, None, None] * tensor[:, :, None]
        Jq = np.sum(Jexp, axis=1)

        pairs = np.array(self._pairs)
        idx = np.where(pairs[:, 0] == pairs[:, 1])
        Jq[idx] /= 2

        return Jq

    def Hq(self, kpoints, anisotropic=True, u=uz):
        """
        Compute the magnon Hamiltonian in reciprocal space.

        Parameters
        ----------
        kpoints : array_like
            k-points at which to evaluate the Hamiltonian
        anisotropic : bool, optional
            Whether to include anisotropic interactions, default True
        u : array_like, optional
            Reference direction for spin quantization axis

        Returns
        -------
        numpy.ndarray
            Magnon Hamiltonian matrix at each k-point
        """
        if self.collinear:
            magmoms = np.zeros((self._index_spin.size, 3))
            magmoms[:, 2] = self.magmoms[self._index_spin]
        else:
            magmoms = self.magmoms[self._index_spin]
        magmoms /= np.linalg.norm(magmoms, axis=-1)[:, None]

        U, V = get_rotation_arrays(magmoms, u=u)

        J0 = self.Jq(np.zeros((1, 3)), anisotropic=anisotropic)
        J0 = -Hermitize(J0)[:, :, 0]
        Jq = -Hermitize(self.Jq(kpoints, anisotropic=anisotropic))

        C = np.diag(np.einsum("ix,ijxy,jy->i", V, 2 * J0, V))
        B = np.einsum("ix,ijkxy,jy->kij", U, Jq, U)
        A1 = np.einsum("ix,ijkxy,jy->kij", U, Jq, U.conj())
        A2 = np.einsum("ix,ijkxy,jy->kij", U.conj(), Jq, U)

        return np.block([[A1 - C, B], [B.swapaxes(-1, -2).conj(), A2 - C]])

    def _magnon_energies(self, kpoints, anisotropic=True, u=uz):
        H = self.Hq(kpoints, anisotropic=anisotropic, u=u)
        n = H.shape[-1] // 2
        I = np.eye(n)

        min_eig = 0.0
        try:
            K = np.linalg.cholesky(H)
        except np.linalg.LinAlgError:
            try:
                K = np.linalg.cholesky(H + 1e-6 * np.eye(2 * n))
            except np.linalg.LinAlgError:
                from warnings import warn

                min_eig = np.min(np.linalg.eigvalsh(H))
                K = np.linalg.cholesky(H - (min_eig - 1e-6) * np.eye(2 * n))
                warn(
                    f"WARNING: The system may be far from the magnetic ground-state. Minimum eigenvalue: {min_eig}. The magnon energies might be unphysical."
                )

        g = np.block([[1 * I, 0 * I], [0 * I, -1 * I]])
        KH = K.swapaxes(-1, -2).conj()

        return np.linalg.eigvalsh(KH @ g @ K)[:, n:] + min_eig

    def get_magnon_bands(
        self,
        kpoints: np.array = np.array([]),
        path: str = None,
        npoints: int = 300,
        special_points: dict = None,
        tol: float = 2e-4,
        pbc: tuple = None,
        cartesian: bool = False,
        labels: list = None,
        anisotropic: bool = True,
        u: np.array = uz,
    ):
        pbc = self._pbc if pbc is None else pbc

        if kpoints.size == 0:
            from ase.cell import Cell

            bandpath = Cell(self._cell).bandpath(
                path=path,
                npoints=npoints,
                special_points=special_points,
                eps=tol,
                pbc=pbc,
            )
            kpoints = bandpath.kpts
            spk = bandpath.special_points
            spk[r"$\Gamma$"] = spk.pop("G", np.zeros(3))
            labels = [
                (i, symbol)
                for symbol in spk
                for i in np.where((kpoints == spk[symbol]).all(axis=1))[0]
            ]
        elif cartesian:
            kpoints = np.linalg.solve(self._cell.T, kpoints.T).T

        bands = self._magnon_energies(kpoints, anisotropic=anisotropic, u=u)

        return labels, bands

    def plot_magnon_bands(self, **kwargs):
        """
        Plot magnon band structure.

        Parameters
        ----------
        **kwargs
            Additional keyword arguments passed to get_magnon_bands and plotting functions
        """
        filename = kwargs.pop("filename", None)
        kpath, bands = self.get_magnon_bands(**kwargs)
        bands_plot = BandsPlot(bands, kpath)
        bands_plot.plot(filename=filename)

    @classmethod
    def load_tb2j(
        cls,
        pickle_file: str = "TB2J.pickle",
        pbc: tuple = (True, True, True),
        anisotropic: bool = False,
        quadratic: bool = False,
    ):
        from pickle import load

        try:
            with open(pickle_file, "rb") as File:
                content = load(File)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"No such file or directory: '{pickle_file}'. Please provide a valid .pickle file."
            )
        else:
            correct_content(content)

        magmoms = content["magmoms"] if content["colinear"] else content["spinat"]
        magnetic_elements = {
            content["atoms"].numbers[i]
            for i, j in enumerate(content["index_spin"])
            if j > -1
        }

        exchange = cls(
            atoms=content["atoms"],
            magmoms=magmoms,
            pbc=pbc,
            collinear=content["colinear"],
            magnetic_elements=magnetic_elements,
        )

        num_pairs = len(exchange.interacting_pairs)
        bkeys = branched_keys(content["distance_dict"].keys(), num_pairs)

        vectors = [
            [content["distance_dict"][key][0] for key in branch] for branch in bkeys
        ]
        exchange.set_vectors(vectors, cartesian=True)
        Jiso = [[content["exchange_Jdict"][key] for key in branch] for branch in bkeys]
        exchange.set_exchange_array("Jiso", Jiso)

        if not content["colinear"] and anisotropic:
            Jani = [[content["Jani_dict"][key] for key in branch] for branch in bkeys]
            exchange.set_exchange_array("Jani", Jani)
            DMI = [[content["dmi_ddict"][key] for key in branch] for branch in bkeys]
            exchange.set_exchange_array("DMI", DMI)
        if quadratic:
            Biquad = [
                [content["biquadratic_Jdict"][key] for key in branch]
                for branch in bkeys
            ]
            exchange.set_exchange_array("Biquad", Biquad)

        return exchange


def plot_tb2j_magnon_bands(
    pickle_file: str = "TB2J.pickle",
    path: str = None,
    npoints: int = 300,
    special_points: dict = None,
    anisotropic: bool = False,
    quadratic: bool = False,
    pbc: tuple = (True, True, True),
    filename: str = None,
):
    """
    Load TB2J data and plot magnon band structure in one step.

    This is a convenience function that combines loading TB2J data and plotting
    magnon bands. It first loads the magnetic structure and exchange interactions
    from a TB2J pickle file, then calculates and plots the magnon band structure.

    Parameters
    ----------
    pickle_file : str, optional
        Path to the TB2J pickle file, default "TB2J.pickle"
    path : str, optional
        High-symmetry k-point path for band structure plot
        (e.g., "GXMG" for a square lattice)
    npoints : int, optional
        Number of k-points for band structure calculation, default 300
    special_points : dict, optional
        Dictionary of special k-points for custom paths
    anisotropic : bool, optional
        Whether to include anisotropic interactions, default False
    quadratic : bool, optional
        Whether to include biquadratic interactions, default False
    pbc : tuple, optional
        Periodic boundary conditions, default (True, True, True)
    filename : str, optional
        If provided, save the plot to this file

    Returns
    -------
    exchange : ExchangeIO
        The ExchangeIO instance containing the loaded data and plot

    Example
    -------
    >>> # Basic usage with default parameters
    >>> plot_tb2j_magnon_bands()

    >>> # Custom path and saving to file
    >>> plot_tb2j_magnon_bands(
    ...     path="GXMG",
    ...     anisotropic=True,
    ...     filename="magnon_bands.png"
    ... )
    """
    # Load the TB2J data
    exchange = ExchangeIO.load_tb2j(
        pickle_file=pickle_file, pbc=pbc, anisotropic=anisotropic, quadratic=quadratic
    )

    # Plot the magnon bands
    exchange.plot_magnon_bands(
        path=path, npoints=npoints, special_points=special_points, filename=filename
    )

    return exchange



import numpy as np
from scipy.special import roots_legendre

from TB2J.utils import simpson_nonuniform_weight


class Contour:
    def __init__(self, emin, emax=0.0):
        self.emin = emin
        self.emax = emax
        self._weights = None

    @property
    def weights(self):
        if self._weights is None:
            self._weights = simpson_nonuniform_weight(self.path)
        return self._weights

    def integrate_func(self, func):
        """
        integrate f along the path
        """
        return np.dot(func(self.path), self.weights)

    def integrate_values(self, values):
        """
        integrate f along the path
        """
        ret = 0
        for i in range(len(values)):
            ret += values[i] * self.weights[i]
        return ret

    def build_path_semicircle(self, npoints, endpoint=True):
        R = (self.emax - self.emin) / 2.0
        R0 = (self.emin + self.emax) / 2.0
        phi = np.linspace(np.pi, 0, num=npoints + 1, endpoint=endpoint)
        p = R0 + R * np.exp(1.0j * phi)
        if endpoint:
            self.path = p
            self.de = np.diff(p)
        else:
            self.path = (p[:-1] + p[1:]) / 2
            self.de = p[1:] - p[:-1]

    def build_path_legendre(self, npoints, endpoint=True):
        p = 13
        x, w = roots_legendre(npoints)
        R = (self.emax - self.emin) / 2.0
        R0 = (self.emin + self.emax) / 2.0
        y1 = -np.log(1 + np.pi * p)
        y2 = 0
        y = (y2 - y1) / 2 * x + (y2 + y1) / 2
        phi = (np.exp(-y) - 1) / p
        path = R0 + R * np.exp(1.0j * phi)
        # weight= -(y2-y1)/2*np.exp(-y)/p*1j*(path-R0)*w
        if endpoint:
            self.path = path
            self.de = np.diff(path)
        else:
            self.path = (path[:-1] + path[1:]) / 2
            self.de = path[1:] - path[:-1]

    def build_path_rectangle(self, height=0.1, nz1=50, nz2=200, nz3=50):
        """
        prepare list of energy for integration.
        The path has three segments:
         emin --1-> emin + 1j*height --2-> emax+1j*height --3-> emax
        """
        nz1, nz2, nz3 = nz1, nz2, nz3
        nz = nz1 + nz2 + nz3
        p = np.zeros(nz + 1, dtype="complex128")
        p[:nz1] = self.emin + np.linspace(0, height, nz1, endpoint=False) * 1j
        p[nz1 : nz1 + nz2] = (
            self.emin
            + height * 1j
            + np.linspace(0, self.emax - self.emin, nz2, endpoint=False)
        )
        p[nz1 + nz2 : nz] = (
            self.emax + height * 1j + np.linspace(0, -height, nz3, endpoint=False) * 1j
        )
        p[-1] = self.emax  # emax
        self.path = (p[:-1] + p[1:]) / 2
        self.de = p[1:] - p[:-1]

    @property
    def npoints(self):
        return len(self.path)

    def de(self):
        return self.de

    def plot(self, ax=None):
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()
        plt.plot(self.path.real, self.path.imag, marker=".")
        plt.show()


def test():
    ct = Contour(emin=-16, emax=0)
    # ct.build_path_semicircle(npoints=100)
    # ct.build_path_rectangle()
    ct.build_path_legendre(npoints=50)
    print(ct.npoints)
    ct.plot()


if __name__ == "__main__":
    test()



import matplotlib.pyplot as plt
import numpy as np


class BandsPlot:
    _UNITS = "meV"
    _NSYSTEMS = 1

    def __init__(self, bands, kpath, **kwargs):
        self.bands = bands
        self.kpath = kpath
        self.bands *= 1000

        plot_options = kwargs
        self.linewidth = plot_options.pop("linewidth", 1.5)
        self.color = plot_options.pop("color", "blue")
        self.fontsize = plot_options.pop("fontsize", 12)
        self.ticksize = plot_options.pop("ticksize", 10)
        self.plot_options = plot_options

    def plot(self, filename=None):
        fig, axs = plt.subplots(1, self._NSYSTEMS, constrained_layout=True)

        kdata = np.arange(self.bands.shape[0])
        for band in self.bands.T:
            axs.plot(
                kdata,
                band,
                linewidth=self.linewidth,
                color=self.color,
                **self.plot_options,
            )

        bmin, bmax = self.bands.min(), self.bands.max()
        ymin, ymax = (
            bmin - 0.05 * np.abs(bmin - bmax),
            bmax + 0.05 * np.abs(bmax - bmin),
        )

        axs.set_ylim([ymin, ymax])
        axs.set_xlim([0, kdata[-1]])

        kpoint_labels = list(zip(*self.kpath))
        axs.set_xticks(*kpoint_labels, fontsize=self.ticksize)
        axs.vlines(
            x=kpoint_labels[0],
            ymin=ymin,
            ymax=ymax,
            color="black",
            linewidth=self.linewidth / 5,
        )

        axs.set_ylabel(f"Energy ({self._UNITS})", fontsize=self.fontsize)

        if filename is None:
            plt.show()
        else:
            fig.save(filename, dpi=300, bbox_inches="tight")



from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import tqdm

# from TB2J.abacus.abacus_wrapper import AbacusSplitSOCParser
from HamiltonIO.abacus.abacus_wrapper import AbacusSplitSOCParser
from HamiltonIO.model.occupations import Occupations
from HamiltonIO.siesta import SislParser
from scipy.linalg import eigh

from TB2J.contour import Contour
from TB2J.green import TBGreen

# from HamiltonIO.model.rotate_spin import rotate_Matrix_from_z_to_axis, rotate_Matrix_from_z_to_sperical
from TB2J.kpoints import monkhorst_pack
from TB2J.mathutils.kR_convert import R_to_k

# from TB2J.abacus.abacus_wrapper import AbacusWrapper, AbacusParser
from TB2J.mathutils.rotate_spin import (
    rotate_spinor_matrix,
)


def get_occupation(evals, kweights, nel, width=0.1):
    occ = Occupations(nel=nel, width=width, wk=kweights, nspin=2)
    return occ.occupy(evals)


def get_density_matrix(evals=None, evecs=None, kweights=None, nel=None, width=0.1):
    occ = get_occupation(evals, kweights, nel, width=width)
    rho = np.einsum("kib, kb, kjb -> kij", evecs, occ, evecs.conj())
    return rho


class MAE:
    def __init__(
        self,
        model,
        kmesh=None,
        gamma=True,
        kpts=None,
        kweights=None,
        width=0.1,
        nel=None,
    ):
        self.model = model
        if nel is not None:
            self.model.nel = nel
        if kpts is None:
            self.kpts = monkhorst_pack(kmesh, gamma_center=gamma)
            self.kweights = np.ones(len(self.kpts), dtype=float) / len(self.kpts)
        else:
            self.kpts = kpts
            self.kweights = kweights
        self.width = width

    def get_band_energy(self):
        evals, evecs = self.model.solve_all(self.kpts)
        occ = get_occupation(evals, self.kweights, self.model.nel, width=self.width)
        eband = np.sum(evals * occ * self.kweights[:, np.newaxis])
        return eband

    def calc_ref(self):
        # calculate the Hk_ref, Sk_ref, Hk_soc_ref, and rho_ref
        self.Sk_ref = R_to_k(self.kpts, self.model.Rlist, self.model.SR)
        self.Hk_xc_ref = R_to_k(self.kpts, self.model.Rlist, self.model._HR_copy)
        self.Hk_soc_ref = R_to_k(self.kpts, self.model.Rlist, self.model.HR_soc)
        self.rho_ref = np.zeros(
            (len(self.kpts), self.model.nbasis, self.model.nbasis), dtype=complex
        )

        evals = np.zeros((len(self.kpts), self.model.nbasis), dtype=float)
        evecs = np.zeros(
            (len(self.kpts), self.model.nbasis, self.model.nbasis), dtype=complex
        )

        for ik, kpt in enumerate(self.kpts):
            # evals, evecs = eigh(self.Hk_xc_ref[ik]+self.Hk_soc_ref[ik], self.Sk_ref[ik])
            evals[ik], evecs[ik] = eigh(self.Hk_xc_ref[ik], self.Sk_ref[ik])
        occ = get_occupation(
            evals, self.kweights, self.model.nel, width=self.model.width
        )
        # occ = fermi(evals, self.model.efermi, width=self.model.width)
        self.rho_ref = np.einsum("kib, kb, kjb -> kij", evecs, occ, evecs.conj())

    def get_band_energy_vs_angles(
        self,
        thetas,
        phis,
    ):
        es = []
        # es2 = []
        # e,rho = self.model.get_band_energy(dm=True)
        # self.calc_ref()
        # thetas = np.linspace(*angle_range, npoints)
        nangles = len(thetas)
        for i in tqdm.trange(nangles):
            theta = thetas[i]
            phi = phis[i]
            self.model.set_Hsoc_rotation_angle([theta, phi])
            e = self.get_band_energy()
            es.append(e)
            # es2.append(e2)
        return es


class MAEGreen(MAE):
    def __init__(
        self,
        model,
        kmesh=None,
        gamma=True,
        kpts=None,
        kweights=None,
        nel=None,
        width=0.1,
        **kwargs,
    ):
        self.model = model
        if nel is not None:
            self.model.nel = nel
        if kpts is None:
            self.kpts = monkhorst_pack(kmesh, gamma_center=gamma)
            self.kweights = np.ones(len(self.kpts), dtype=float) / len(self.kpts)
        else:
            self.kpts = kpts
            self.kweights = kweights
        self.width = width
        model.set_so_strength(0.0)
        evals, evecs = model.solve_all(self.kpts)
        occ = Occupations(
            nel=self.model.nel, width=self.width, wk=self.kweights, nspin=2
        )
        # occ.occupy(evals)
        efermi = occ.efermi(evals)
        print(f"{efermi=}")
        self.G = TBGreen(model, kmesh, efermi=efermi, gamma=gamma, **kwargs)
        self.emin = -52
        self.emax = 0
        self.nz = 100
        self._prepare_elist()

    def _prepare_elist(self, method="legendre"):
        """
        prepare list of energy for integration.
        The path has three segments:
         emin --1-> emin + 1j*height --2-> emax+1j*height --3-> emax
        """
        self.contour = Contour(self.emin, self.emax)
        # if method.lower() == "rectangle":
        #    self.contour.build_path_rectangle(
        #        height=self.height, nz1=self.nz1, nz2=self.nz2, nz3=self.nz3
        #    )
        if method.lower() == "semicircle":
            self.contour.build_path_semicircle(npoints=self.nz, endpoint=True)
        elif method.lower() == "legendre":
            self.contour.build_path_legendre(npoints=self.nz, endpoint=True)
        else:
            raise ValueError(f"The path cannot be of type {method}.")

    def get_efermi(self):
        evals, evecs = self.model.solve_all(self.kpts)
        occ = Occupations(
            nel=self.model.nel, width=self.model.width, wk=self.kweights, nspin=2
        )
        occ.get_efermi(evals, self.kweights)

    def get_perturbed(self, e, thetas, phis):
        G0K = self.G.get_Gk_all(e)
        Hsoc_k = self.model.get_Hk_soc(self.kpts)
        dE_ang = []
        for theta, phi in zip(thetas, phis):
            dE = 0.0
            for i, dHk in enumerate(Hsoc_k):
                dHi = rotate_spinor_matrix(dHk, theta, phi)
                GdH = G0K[i] @ dHi
                # dE += np.trace(GdH @ G0K[i].T.conj() @ dHi) * self.kweights[i]
                dE += np.trace(GdH @ GdH) * self.kweights[i]
            dE_ang.append(dE)
        return np.array(dE_ang)

    def get_band_energy_vs_angles(self, thetas, phis):
        es = np.zeros(len(thetas))
        for ie, e in enumerate(tqdm.tqdm(self.contour.path)):
            dE_angle = self.get_perturbed(e, thetas, phis)
            es += np.imag(dE_angle * self.contour.weights[ie])
        return -es / np.pi / 2


def get_model_energy(model, kmesh, gamma=True):
    ham = MAE(model, kmesh, gamma=gamma)
    return ham.get_band_energy()


def abacus_get_MAE(
    path_nosoc,
    path_soc,
    kmesh,
    thetas,
    phis,
    gamma=True,
    outfile="MAE.txt",
    nel=None,
    width=0.1,
):
    """Get MAE from Abacus with magnetic force theorem. Two calculations are needed. First we do an calculation with SOC but the soc_lambda is set to 0. Save the density. The next calculatin we start with the density from the first calculation and set the SOC prefactor to 1. With the information from the two calcualtions, we can get the band energy with magnetic moments in the direction, specified in two list, thetas, and phis."""
    parser = AbacusSplitSOCParser(
        outpath_nosoc=path_nosoc, outpath_soc=path_soc, binary=False
    )
    model = parser.parse()
    if nel is not None:
        model.nel = nel
    ham = MAEGreen(model, kmesh, gamma=gamma, width=width)
    es = ham.get_band_energy_vs_angles(thetas, phis)
    if outfile:
        with open(outfile, "w") as f:
            f.write("#theta, phi, energy\n")
            for theta, phi, e in zip(thetas, phis, es):
                f.write(f"{theta:5.3f}, {phi:5.3f}, {e:10.9f}\n")
    return es


def siesta_get_MAE(
    fdf_fname, kmesh, thetas, phis, gamma=True, outfile="MAE.txt", nel=None, width=0.1
):
    """ """
    model = SislParser(fdf_fname=fdf_fname, read_H_soc=True).get_model()
    if nel is not None:
        model.nel = nel
    ham = MAEGreen(model, kmesh, gamma=gamma, width=width)
    # es = ham.get_band_energy_vs_angles(thetas, phis)
    es = ham.get_band_energy_vs_angles(thetas, phis)
    if outfile:
        with open(outfile, "w") as f:
            f.write("#theta, psi, energy\n")
            for theta, psi, e in zip(thetas, phis, es):
                # f.write(f"{theta}, {psi}, {e}\n")
                f.write(f"{theta:5.3f}, {psi:5.3f}, {e:10.9f}\n")
    return es


def test_AbacusSplitSOCWrapper():
    # path = Path("~/projects/2D_Fe").expanduser()
    path = Path("~/projects/TB2Jflows/examples/2D_Fe/Fe_z").expanduser()
    outpath_nosoc = f"{path}/soc0/OUT.ABACUS"
    outpath_soc = f"{path}/soc1/OUT.ABACUS"
    parser = AbacusSplitSOCParser(
        outpath_nosoc=outpath_nosoc, outpath_soc=outpath_soc, binary=False
    )
    model = parser.parse()
    kmesh = [6, 6, 1]

    r = MAE(model, kmesh, gamma=True)
    # thetas, es = r.get_band_energy_vs_theta(angle_range=(0, np.pi*2), rotation_axis="z", initial_direction=(1,0,0),  npoints=21)
    thetas, es, es2 = r.get_band_energy_vs_theta(
        angle_range=(0, np.pi),
        rotation_axis="y",
        initial_direction=(0, 0, 1),
        npoints=11,
    )
    # print the table of thetas and es, es2
    for theta, e, e2 in zip(thetas, es, es2):
        print(f"{theta=}, {e=}, {e2=}")

    plt.plot(thetas / np.pi, es - es[0], marker="o")
    plt.plot(thetas / np.pi, es2 - es2[0], marker=".")
    plt.savefig("E_along_z_x_z.png")
    plt.show()


def abacus_get_MAE_cli():
    import argparse

    parser = argparse.ArgumentParser(
        description="Get MAE from Abacus with magnetic force theorem. Two calculations are needed. First we do an calculation with SOC but the soc_lambda is set to 0. Save the density. The next calculatin we start with the density from the first calculation and set the SOC prefactor to 1. With the information from the two calcualtions, we can get the band energy with magnetic moments in the direction, specified in two list, thetas, and phis. "
    )
    parser.add_argument("path_nosoc", type=str, help="Path to the  calculation with ")
    parser.add_argument("path_soc", type=str, help="Path to the SOC calculation")
    parser.add_argument("thetas", type=float, nargs="+", help="Thetas")
    parser.add_argument("psis", type=float, nargs="+", help="Phis")
    parser.add_argument("kmesh", type=int, nargs=3, help="K-mesh")
    parser.add_argument(
        "--gamma", action="store_true", help="Use Gamma centered kpoints"
    )
    parser.add_argument(
        "--outfile",
        type=str,
        help="The angles and the energey will be saved in this file.",
    )
    args = parser.parse_args()
    abacus_get_MAE(
        args.path_nosoc,
        args.path_soc,
        args.kmesh,
        args.thetas,
        args.psis,
        gamma=args.gamma,
        outfile=args.outfile,
    )


if __name__ == "__main__":
    abacus_get_MAE_cli()



from warnings import warn

import numpy as np

valid_symbols = [
    # 0
    "Ae",
    # 1
    "H",
    "He",
    # 2
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    # 3
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    # 4
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    # 5
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    # 6
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    # 7
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og",
]


def get_attribute_array(array, attribute, dtype=float):
    try:
        the_array = np.array(array, dtype=dtype)
    except (IndexError, ValueError, TypeError) as err:
        typename = dtype.__name__
        raise type(err)(
            f"'{attribute}' must be an arraylike object of '{typename}' elements."
        )

    return the_array


def validate_symbols(value):
    try:
        values_list = value.split()
    except AttributeError:
        try:
            values_list = list(value)
        except (ValueError, TypeError):
            raise TypeError("'elements' must be an iterable of 'str' or 'int' entries.")

    if all(isinstance(s, str) for s in values_list):
        symbols = values_list
        for symbol in symbols:
            if symbol not in valid_symbols:
                raise ValueError(f"Unrecognized element '{symbol}'.")
    elif all(isinstance(i, (int, np.integer)) for i in values_list):
        if any(i < 0 for i in values_list):
            raise ValueError("Atomic numbers must be positive.")
        try:
            symbols = [valid_symbols[i] for i in values_list]
        except IndexError:
            raise ValueError("Atomic number exceeds 118.")
    else:
        raise ValueError("'elements' must be an iterable of 'str' or 'int' entries.")

    return symbols


class BaseMagneticStructure:
    def __init__(
        self,
        atoms=None,
        cell=None,
        elements=None,
        positions=None,
        magmoms=None,
        pbc=None,
        collinear=True,
    ):
        if atoms is not None:
            if any(arg is not None for arg in [cell, elements, positions]):
                warn(
                    "WARNING: 'atoms' overrides the 'cell', 'elements', and 'positions' arguments."
                )
            cell = atoms.cell.array
            elements = atoms.numbers
            positions = atoms.get_scaled_positions()
            pbc = atoms.pbc if pbc is None else pbc
        else:
            if cell is None:
                cell = np.zeros((3, 3))
            if elements is None:
                elements = ()
            if positions is None:
                positions = np.zeros((len(elements), 3))
            if pbc is None:
                pbc = (True, True, True)
        if magmoms is None:
            magmoms_shape = positions.shape[0] if collinear else positions.shape
            magmoms = np.zeros(magmoms_shape)

        self.cell = cell
        self.elements = elements
        self.positions = positions
        self.collinear = collinear
        self.magmoms = magmoms
        self.pbc = pbc

        self._Q = None
        self._n = np.array([0, 0, 1])

    @property
    def cell(self):
        return self._cell

    @property
    def reciprocal_cell(self):
        return 2 * np.pi * np.linalg.inv(self._cell)

    @cell.setter
    def cell(self, value):
        cell_array = get_attribute_array(value, "cell")
        if cell_array.shape != (3, 3):
            raise ValueError("'cell' must have a (3, 3) shape.")

        self._cell = cell_array

    @property
    def elements(self):
        return self._symbols

    @elements.setter
    def elements(self, value):
        symbols = validate_symbols(value)
        self._symbols = symbols

    @property
    def numbers(self):
        return [valid_symbols.index(symbol) for symbol in self._symbols]

    @property
    def positions(self):
        return self._positions

    @property
    def cartesian_positions(self):
        return self._positions @ self._cell

    @positions.setter
    def positions(self, value):
        natoms = len(self.elements)
        posarray = get_attribute_array(value, "positions")
        if posarray.shape != (natoms, 3):
            raise ValueError(
                "'positions' must have a (natoms, 3) shape, where natoms is the length of 'elements'. Make sure to set 'elements' first."
            )

        self._positions = posarray

    @property
    def magmoms(self):
        return self._magmoms

    @magmoms.setter
    def magmoms(self, value):
        magarray = get_attribute_array(value, "magmoms")
        if self.collinear and magarray.shape != (len(self._positions),):
            raise ValueError(
                "'magmoms' must be a 1dim array with the same length as 'positions'. Make sure to set 'positions' first."
            )
        elif not self.collinear and magarray.shape != self._positions.shape:
            raise ValueError(
                "'magmoms' must have the same shape as 'positions'. Make sure to set 'positions' first."
            )

        self._magmoms = magarray

    @property
    def pbc(self):
        return self._pbc

    @pbc.setter
    def pbc(self, value):
        try:
            periodic_boundary_conditions = tuple(bool(b) for b in value)
        except (ValueError, TypeError):
            raise TypeError("'pbc' must be an iterable of 'bool' entries.")

        if len(periodic_boundary_conditions) != 3:
            raise ValueError("'pbc' must be of length 3.")

        self._pbc = periodic_boundary_conditions

    @property
    def propagation_vector(self):
        return self._Q

    @propagation_vector.setter
    def propagation_vector(self, value):
        Q = get_attribute_array(value, "propagation_vector")
        if Q.shape != (3,):
            raise ValueError(
                "The 'propagation_vector' must be a vector with 3 numbers."
            )

        self._Q = Q

    @property
    def normal_vector(self):
        return self._n

    @normal_vector.setter
    def normal_vector(self, value):
        n = get_attribute_array(value, "normal_vector")
        if n.shape != (3,):
            raise ValueError("The 'normal_vector' must be a vector with 3 numbers.")

        self._n = n

    def to_ase(self):
        from ase.atoms import Atoms

        atoms = Atoms(
            cell=self.cell,
            positions=self.cartesian_positions,
            numbers=[valid_symbols.index(symbol) for symbol in self.elements],
            pbc=self.pbc,
        )

        return atoms

    @classmethod
    def from_structure_file(
        cls, filename, magmoms=None, collinear=True, pbc=(True, True, True)
    ):
        from ase.io import read

        atoms = read(filename)
        atoms.pbc = pbc
        obj = cls(atoms=atoms, magmoms=magmoms, collinear=collinear)

        return obj



#!/usr/bin/env python
from setuptools import find_packages, setup

__version__ = "0.9.9.8"

long_description = """TB2J is a Python package aimed to compute automatically the magnetic interactions (superexchange  and Dzyaloshinskii-Moriya) between atoms of magnetic crystals from DFT Hamiltonian based on Wannier functions or Linear combination of atomic orbitals. It uses the Green's function method and take the local rigid spin rotation as a perturbation. The package can take the output from Wannier90, which is interfaced with many density functional theory codes or from codes based on localised orbitals. A minimal user input is needed, which allows for an easily integration into a high-throughput workflows. """

setup(
    name="TB2J",
    version=__version__,
    description="TB2J: First principle to Heisenberg exchange J using tight-binding Green function method",
    long_description=long_description,
    author="Xu He",
    author_email="mailhexu@gmail.com",
    license="BSD-2-clause",
    packages=find_packages(),
    scripts=[
        "scripts/wann2J.py",
        "scripts/siesta2J.py",
        "scripts/abacus2J.py",
        "scripts/TB2J_rotate.py",
        "scripts/TB2J_rotateDM.py",
        "scripts/TB2J_merge.py",
        "scripts/TB2J_magnon.py",
        "scripts/TB2J_magnon2.py",
        "scripts/TB2J_magnon_dos.py",
        "scripts/TB2J_downfold.py",
        "scripts/TB2J_eigen.py",
    ],
    entry_points={
        "console_scripts": [
            "TB2J_symmetrize.py=TB2J.symmetrize_J:symmetrize_J_cli",
            "lawaf2J.py=TB2J.interfaces.lawaf_interface:lawaf2J_cli",
        ]
    },
    install_requires=[
        "numpy<2.0",
        "scipy",
        "matplotlib",
        "ase>=3.19",
        "tqdm",
        "pathos",
        "packaging>=20.0",
        "HamiltonIO>=0.2.3",
        "pre-commit",
        "sympair>0.1.0",
        "sisl>=0.9.0",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: BSD License",
    ],
    python_requires=">=3.6",
)



"""
IO module. called io_xml for historical reason.
write not only xml output.
- xml
- multibinit input template
- vampire inputs
- uppasd inputs (not tested yet)
- tom's asd inputs.
"""

# matplotlib.use("Agg")
import gc
import os
import pickle
from collections.abc import Iterable
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np

from TB2J import __version__
from TB2J.io_exchange.io_txt import write_Jq_info
from TB2J.Jtensor import combine_J_tensor
from TB2J.kpoints import monkhorst_pack
from TB2J.spinham.spin_api import SpinModel
from TB2J.utils import symbol_number


class SpinIO(object):
    def __init__(
        self,
        atoms,
        spinat,
        charges,
        index_spin,
        orbital_names={},
        colinear=True,
        distance_dict=None,
        exchange_Jdict=None,
        Jiso_orb=None,
        DMI_orb=None,
        Jani_orb=None,
        dJdx=None,
        dJdx2=None,
        dmi_ddict=None,
        Jani_dict=None,
        biquadratic_Jdict=None,
        debug_dict=None,
        k1=None,
        k1dir=None,
        NJT_Jdict=None,
        NJT_ddict=None,
        damping=None,
        gyro_ratio=None,
        write_experimental=True,
        description=None,
    ):
        """
        :param atoms: Ase atoms structure.
        :param spinat: spin for each atom. (3*natom)
        :param charges: charges for each atom (natom)
        :param index_spin: index of spin in the spin potential for each atom.
        :param colinear: whether the parameters are for a colinear wannier calculation.
        :param distance_dict: {(R, i,j ): distance} , here i and j are spin indices
        :param exchange_Jdict: {(R, i,j): J}
        :param Jiso_orb: {(R, i,j): J_orb}
        :param DMI_orb: {(R, i,j): D_orb}
        :param Jani_orb: {(R, i,j): Jani_orb}
        :param dJdx: {(R, i,j): dJdx}
        :param dJdx2: {(R, i,j): dJdx2}
        :param dmi_ddict:{(R, i,j): DMI}
        :param Jani_dict: {(R, i,j): Jani'}, Jani is a 3*3 matrix
        :param biqudratic_Jdict: {(R, i,j ): J}
        :param k1: single ion anisotropy amplitude.o
        :param kdir: directio of k1
        :param NJT_Jdict: exhange calculated using NJT method
        :param NJT_ddict: DMI calculated using NJT method
        :param damping: damping factor
        :param gyro_ratio:  gyromagnetic ratio
        :param write_experimental: write_experimental data to output files
        :param description: add some description into the xml file.
        """
        self.atoms = atoms  #: atomic structures, ase.Atoms object
        self.index_spin = index_spin
        #: index of spin linked to atoms. -1 if non-magnetic
        self.spinat = spinat  #: spin for each atom. shape of (natom, 3)
        self.colinear = colinear  #: If the calculation is collinear or not
        if (
            self.colinear
            and isinstance(self.spinat, np.ndarray)
            and self.spinat.shape[1] == 3
        ):
            self.magmoms = np.array(self.spinat)[:, 2]
        self.charges = charges
        #: A dictionary of distances, the keys are (i,j, R),
        # where i and j are spin index and R is the cell index,
        # a tuple of three integers.
        self.distance_dict = distance_dict
        self._ind_atoms = {}  #: The index of atom for each spin.
        for iatom, ispin in enumerate(self.index_spin):
            if ispin >= 0:
                self._ind_atoms[ispin] = iatom

        if exchange_Jdict is not None:
            self.has_exchange = True  #: whether there is isotropic exchange
            #: The dictionary of :math:`J_{ij}(R)`, the keys are (i,j, R),
            # where R is a tuple, and the value is the isotropic exchange
            self.exchange_Jdict = exchange_Jdict
        else:
            self.has_exchange = False
            self.exchange_Jdict = None

        self.Jiso_orb = Jiso_orb

        self.DMI_orb = DMI_orb
        self.Jani_orb = Jani_orb

        self.dJdx = dJdx
        self.dJdx2 = dJdx2

        if dmi_ddict is not None:
            self.has_dmi = True  #: Whether there is DMI.
            #: The dictionary of DMI. the key is the same as exchange_Jdict, the values are 3-d vectors (Dx, Dy, Dz).
            self.dmi_ddict = dmi_ddict
        else:
            self.has_dmi = False
            self.dmi_ddict = None

        if Jani_dict is not None:
            self.has_bilinear = True  #: Whether there is anisotropic exchange term
            #: The dictionary of anisotropic exchange. The vlaues are matrices of shape (3,3).
            self.Jani_dict = Jani_dict
        else:
            self.has_bilinear = False
            self.Jani_dict = None

        if k1 is not None and k1dir is not None:
            self.has_uniaxial_anistropy = True
            self.k1 = k1
            self.k1dir = k1dir
        else:
            self.has_uniaxial_anistropy = False
            self.k1 = None
            self.k1dir = None

        self.has_bilinear = not (Jani_dict == {} or Jani_dict is None)

        self.has_biquadratic = not (
            biquadratic_Jdict == {} or biquadratic_Jdict is None
        )
        self.biquadratic_Jdict = biquadratic_Jdict

        if NJT_ddict is not None:
            self.has_NJT_dmi = True
        self.NJT_ddict = NJT_ddict

        if NJT_Jdict is not None:
            self.has_NJT_exchange = True
        self.NJT_Jdict = NJT_Jdict

        natom = len(self.atoms)
        if gyro_ratio is None:
            self.gyro_ratio = [1.0] * natom  #: Gyromagnetic ratio for each atom
        elif isinstance(gyro_ratio, Iterable):
            self.gyro_ratio = gyro_ratio
        else:
            self.gyro_ratio = [gyro_ratio] * natom

        if damping is None:
            self.damping = [1.0] * natom  # damping factor for each atom
        elif isinstance(damping, Iterable):
            self.damping = damping
        else:
            self.damping = [damping] * natom

        self.debug_dict = debug_dict

        self.write_experimental = write_experimental

        now = datetime.now()
        self.description = f"""Exchange parameters generated by TB2J {__version__}
Generation time: {now.strftime("%y/%m/%d %H:%M:%S")}
"""
        if description is not None:
            self.description += description

        self.orbital_names = orbital_names
        self.TB2J_version = __version__

    def _build_Rlist(self):
        Rset = set()
        ispin_set = set()
        for R, i, j in self.exchange_Jdict:
            Rset.add(R)
            ispin_set.add(i)
            ispin_set.add(j)
        self.Rlist = list(Rset)
        self.ispin_list = list(ispin_set)
        self.nspin = len(self.ispin_list)
        assert self.nspin == max(self.ispin_list) + 1

    def _build_ind_atoms(self):
        self._ind_atoms = {}  #: The index of atom for each spin.
        for iatom, ispin in enumerate(self.index_spin):
            if ispin >= 0:
                self._ind_atoms[ispin] = iatom

    @property
    def ind_atoms(self):
        if not self._ind_atoms:
            self._build_ind_atoms()
        return self._ind_atoms

    def iatom(self, i):
        return self.ind_atoms[i]

    def get_spin_ispin(self, i):
        return self.spinat[self.iatom(i)]

    def get_symbol_number_ispin(self, symnum):
        """
        Return the spin index for a given symbol number.
        """
        symdict = symbol_number(self.atoms)
        return self.index_spin[symdict[symnum]]

    def i_spin(self, i):
        if isinstance(i, int):
            return i
        elif isinstance(i, str):
            return self.get_symbol_number_ispin(i)
        else:
            raise ValueError("i must be either an integer or a string.")

    def get_charge_ispin(self, i):
        i = self.i_spin(i)
        return self.charges[self.iatom(i)]

    def get_spin_iatom(self, iatom):
        return self.spinat[iatom]

    def get_charge_iatom(self, iatom):
        return self.charges[iatom]

    def ijR_index_spin_to_atom(self, i, j, R):
        return (self.iatom(i), self.iatom(j), R)

    def ijR_index_atom_to_spin(self, iatom, jatom, R):
        return (self.index_spin[iatom], self.index_spin[jatom], R)

    def ijR_list(self):
        return [(i, j, R) for R, i, j in self.exchange_Jdict]

    def ijR_list_index_atom(self):
        return [self.ijR_index_spin_to_atom(i, j, R) for R, i, j in self.exchange_Jdict]

    def get_J(self, i, j, R, default=None):
        i = self.i_spin(i)
        j = self.i_spin(j)
        key = (
            tuple(R),
            i,
            j,
        )
        if self.exchange_Jdict is not None and key in self.exchange_Jdict:
            return self.exchange_Jdict[key]
        else:
            return default

    def get_Jiso(self, i, j, R, default=None):
        i = self.i_spin(i)
        j = self.i_spin(j)
        key = (
            tuple(R),
            i,
            j,
        )
        if self.exchange_Jdict is not None and key in self.exchange_Jdict:
            return self.exchange_Jdict[key]
        else:
            return default

    def get_DMI(self, i, j, R, default=None):
        i = self.i_spin(i)
        j = self.i_spin(j)
        key = (
            tuple(R),
            i,
            j,
        )
        if self.dmi_ddict is not None and key in self.dmi_ddict:
            return self.dmi_ddict[(tuple(R), i, j)]
        else:
            return default

    def get_Jani(self, i, j, R, default=None):
        """
        Return the anisotropic exchange tensor for atom i and j, and cell R.
        param i : spin index i
        param j: spin index j
        param R (tuple of integers): cell index R
        """
        i = self.i_spin(i)
        j = self.i_spin(j)
        key = (
            tuple(R),
            i,
            j,
        )
        if self.Jani_dict is not None and key in self.Jani_dict:
            return self.Jani_dict[(tuple(R), i, j)]
        else:
            return default

    def get_J_tensor(self, i, j, R, iso_only=False):
        """
        Return the full exchange tensor for atom i and j, and cell R.
        param i : spin index i
        param j: spin index j
        param R (tuple of integers): cell index R
        """
        i = self.i_spin(i)
        j = self.i_spin(j)
        if iso_only:
            J = self.get_Jiso(i, j, R)
            if J is not None:
                Jtensor = np.eye(3) * self.get_J(i, j, R)
            else:
                Jtensor = np.eye(3) * 0
        else:
            Jtensor = combine_J_tensor(
                Jiso=self.get_J(i, j, R),
                D=self.get_DMI(i, j, R),
                Jani=self.get_Jani(i, j, R),
            )
        return Jtensor

    def get_full_Jtensor_for_one_R_i3j3(self, R, iso_only=False):
        """
        Return the full exchange tensor of all i and j for cell R.
        param R (tuple of integers): cell index R
        returns:
            Jmat: (3*nspin,3*nspin) matrix.
        """
        n3 = self.nspin * 3
        Jmat = np.zeros((n3, n3), dtype=float)
        for i in range(self.nspin):
            for j in range(self.nspin):
                Jmat[i * 3 : i * 3 + 3, j * 3 : j * 3 + 3] = self.get_J_tensor(
                    i, j, R, iso_only=iso_only
                )
        return Jmat

    def get_full_Jtensor_for_one_R_ij33(self, R, iso_only=False):
        """
        Return the full exchange tensor of all i and j for cell R.
        param R (tuple of integers): cell index R
        returns:
            Jmat: (nspin,nspin,3,3) matrix.
        """
        n = self.nspin
        Jmat = np.zeros((n, n, 3, 3), dtype=float)
        for i in range(self.nspin):
            for j in range(self.nspin):
                Jmat[i, j, :, :] = self.get_J_tensor(i, j, R, iso_only=iso_only)
        return Jmat

    def get_full_Jtensor_for_Rlist(self, asr=False, iso_only=True, order="i3j3"):
        n = self.nspin
        n3 = n * 3
        nR = len(self.Rlist)
        if order == "i3j3":
            Jmat = np.zeros((nR, n3, n3), dtype=float)
            for iR, R in enumerate(self.Rlist):
                Jmat[iR] = self.get_full_Jtensor_for_one_R_i3j3(R, iso_only=iso_only)
            if asr:
                iR0 = np.argmin(np.linalg.norm(self.Rlist, axis=1))
                assert np.linalg.norm(self.Rlist[iR0]) == 0
                for i in range(n3):
                    sum_JRi = np.sum(np.sum(Jmat, axis=0)[i])
                    Jmat[iR0][i, i] -= sum_JRi

        elif order == "ij33":
            Jmat = np.zeros((nR, n, n, 3, 3), dtype=float)
            Jmat[iR] = self.get_full_Jtensor_for_one_R_ij33(R, iso_only=iso_only)
            if asr:
                iR0 = np.argmin(np.linalg.norm(self.Rlist, axis=1))
                assert np.linalg.norm(self.Rlist[iR0]) == 0
                for i in range(n):
                    sum_JRi = np.sum(np.sum(Jmat, axis=0)[i])
                    Jmat[iR0][i, i] -= sum_JRi
        else:
            raise ValueError("order must be either 'i3j3' or 'ij33'.")
        return Jmat

    def write_pickle(self, path="TB2J_results", fname="TB2J.pickle"):
        if not os.path.exists(path):
            os.makedirs(path)
        fname = os.path.join(path, fname)
        with open(fname, "wb") as myfile:
            try:
                pickle.dump(self.__dict__, myfile)
            except Exception as ex:
                print(f"Pickle not written due to {ex}")

    @classmethod
    def load_pickle(cls, path="TB2J_results", fname="TB2J.pickle"):
        fname = os.path.join(path, fname)
        with open(fname, "rb") as myfile:
            d = pickle.load(myfile)
        obj = cls(
            atoms=d["atoms"],
            spinat=d["spinat"],
            charges=d["charges"],
            index_spin=d["index_spin"],
        )
        obj.__dict__.update(d)
        obj._build_Rlist()
        return obj

    def write_all(self, path="TB2J_results"):
        self.write_pickle(path=path)
        self.write_txt(path=path)
        if self.Jiso_orb:
            self.write_txt(
                path=path,
                fname="exchange_orb_decomposition.out",
                write_orb_decomposition=True,
            )
        self.write_multibinit(path=os.path.join(path, "Multibinit"))
        self.write_tom_format(path=os.path.join(path, "TomASD"))
        self.write_vampire(path=os.path.join(path, "Vampire"))

        self.plot_all(savefile=os.path.join(path, "JvsR.pdf"))
        # self.write_Jq(kmesh=[9, 9, 9], path=path)

    def write_txt(self, *args, **kwargs):
        from TB2J.io_exchange.io_txt import write_txt

        write_txt(self, *args, **kwargs)

    # def write_txt_with_orb(self, path):
    #    from TB2J.io_exchange.io_txt import write_txt
    #    write_txt_with_orb(
    #        self, path=path, write_experimental=self.write_experimental)

    def write_multibinit(self, path):
        from TB2J.io_exchange.io_multibinit import write_multibinit

        write_multibinit(self, path=path)

    def write_Jq(self, kmesh, path, gamma=True, output_fname="EigenJq.txt", **kwargs):
        m = SpinModel(fname=os.path.join(path, "Multibinit", "exchange.xml"))
        m.set_ham(**kwargs)
        kpts = monkhorst_pack(kmesh, gamma_center=gamma)

        evals, evecs = m.ham.solve_k(kpts, Jq=True)
        with open(os.path.join(path, output_fname), "w") as myfile:
            myfile.write("=" * 60)
            myfile.write("\n")
            myfile.write("Generated by TB2J %s.\n" % (__version__))
            myfile.write("=" * 60 + "\n")
            myfile.write(
                "The spin ground state is estimated by calculating\n the eigenvalues and eigen vectors of J(q):\n"
            )
            write_Jq_info(self, kpts, evals, evecs, myfile, special_kpoints={})

    def model(self, path):
        m = SpinModel(fname=os.path.join(path, "Multibinit", "exchange.xml"))
        return m

    def plot_JvsR(
        self,
        ax=None,
        color="blue",
        marker="o",
        fname=None,
        show=False,
        **kwargs,
    ):
        if ax is None:
            fig, ax = plt.subplots()
        ds = []
        Js = []
        for key, val in self.exchange_Jdict.items():
            d = self.distance_dict[key][1]
            ds.append(d)
            Js.append(val * 1e3)
        ax.scatter(ds, Js, marker=marker, color=color, **kwargs)
        ax.axhline(color="gray")
        ax.set_xlabel("Distance ($\AA$)")
        ax.set_ylabel("J (meV)")
        if fname is not None:
            plt.savefig(fname)
        if show:
            plt.show()
        return ax

    def plot_DvsR(self, ax=None, fname=None, show=False):
        if ax is None:
            fig, ax = plt.subplots()
        ds = []
        Ds = []
        for key, val in self.dmi_ddict.items():
            d = self.distance_dict[key][1]
            ds.append(d)
            Ds.append(val * 1e3)
        Ds = np.array(Ds)
        ax.scatter(ds, Ds[:, 0], marker="s", color="r", label="Dx")
        ax.scatter(
            ds, Ds[:, 1], marker="o", edgecolors="g", facecolors="none", label="Dy"
        )
        ax.scatter(
            ds, Ds[:, 2], marker="D", edgecolors="b", facecolors="none", label="Dz"
        )
        ax.axhline(color="gray")
        ax.legend(loc=1)
        ax.set_ylabel("D (meV)")
        ax.set_xlabel("Distance ($\AA$)")
        if fname is not None:
            plt.savefig(fname)
        if show:
            plt.show()
        return ax

    def plot_JanivsR(self, ax=None, fname=None, show=False):
        if ax is None:
            fig, ax = plt.subplots()
        ds = []
        Jani = []
        for key, val in self.Jani_dict.items():
            d = self.distance_dict[key][1]
            ds.append(d)
            # val = val - np.diag([np.trace(val) / 3] * 3)
            Jani.append(val * 1e3)
        Jani = np.array(Jani)
        s = "xyz"
        for i in range(3):
            ax.scatter(ds, Jani[:, i, i], marker="s", label=f"J{s[i]}{s[i]}")
        c = "rgb"
        for ic, (i, j) in enumerate([(0, 1), (0, 2), (1, 2)]):
            ax.scatter(
                ds,
                Jani[:, i, j],
                edgecolors=c[ic],
                facecolors="none",
                label=f"J{s[i]}{s[j]}",
            )
        ax.axhline(color="gray")
        ax.legend(loc=1, ncol=2)
        ax.set_xlabel("Distance ($\AA$)")
        ax.set_ylabel("Jani (meV)")
        if fname is not None:
            plt.savefig(fname)
        if show:
            plt.show()
        return ax

    def plot_all(self, title=None, savefile=None, show=False):
        if self.has_dmi and self.has_bilinear:
            naxis = 3
        else:
            naxis = 1
        fig, axes = plt.subplots(naxis, 1, sharex=True, figsize=(5, 2.2 * naxis))

        if self.has_dmi and self.has_bilinear:
            self.plot_JvsR(axes[0])
            self.plot_DvsR(axes[1])
            self.plot_JanivsR(axes[2])
        else:
            self.plot_JvsR(axes)

        if title is not None:
            fig.suptitle(title)
        plt.tight_layout()
        plt.subplots_adjust(hspace=0.25, top=0.93)
        if savefile is not None:
            plt.savefig(savefile)
        if show:
            plt.show()
        plt.clf()
        plt.close()
        gc.collect()  # This is to fix the tk error if multiprocess is used.
        return fig, axes

    def write_tom_format(self, path):
        from TB2J.io_exchange.io_tomsasd import write_tom_format

        write_tom_format(self, path=path)

    def write_vampire(self, path):
        from TB2J.io_exchange.io_vampire import write_vampire

        write_vampire(self, path=path)

    def write_uppasd(self, path):
        from TB2J.io_exchange.io_uppasd import write_uppasd

        write_uppasd(self, path=path)


def gen_distance_dict(ind_mag_atoms, atoms, Rlist):
    distance_dict = {}
    ind_matoms = ind_mag_atoms
    for ispin, iatom in enumerate(ind_matoms):
        for jspin, jatom in enumerate(ind_matoms):
            for R in Rlist:
                pos_i = atoms.get_positions()[iatom]
                pos_jR = atoms.get_positions()[jatom] + np.dot(R, atoms.get_cell())
                vec = pos_jR - pos_i
                distance = np.sqrt(np.sum(vec**2))
                distance_dict[(tuple(R), ispin, jspin)] = (vec, distance)
    return distance_dict


def test_spin_io():
    import numpy as np
    from ase import Atoms

    atoms = Atoms(
        "SrMnO3",
        cell=np.eye(3) * 3.8,
        scaled_positions=[
            [0, 0, 0],
            [0.5, 0.5, 0.5],
            [0, 0.5, 0.5],
            [0.5, 0, 0.5],
            [0.5, 0.5, 0],
        ],
    )
    spinat = [[0, 0, x] for x in [0, 3, 0, 0, 0]]
    charges = [2, 4, 5, 5, 5]
    index_spin = [-1, 0, -1, -1, -1]
    Rlist = [[0, 0, 0], [0, 0, 1]]
    ind_mag_atoms = [1]
    distance_dict = gen_distance_dict(ind_mag_atoms, atoms, Rlist)

    R0 = (0, 0, 0)
    R1 = (0, 0, 1)
    exchange_Jdict = {(R0, 0, 0): 1.2, (R1, 0, 0): 1.1}

    sio = SpinIO(
        atoms,
        spinat,
        charges,
        index_spin,
        colinear=True,
        distance_dict=distance_dict,
        exchange_Jdict=exchange_Jdict,
    )

    sio.write_all()


if __name__ == "__main__":
    test_spin_io()



import numpy as np
from scipy.linalg import eigh


def Lowdin(S):
    """
    Calculate S^(-1/2).
    Which is used in lowind's symmetric orthonormalization.
    psi_prime = S^(-1/2) psi
    """
    eigval, eigvec = eigh(S)
    S_half = eigvec @ np.diag(np.sqrt(1.0 / eigval)) @ (eigvec.T.conj())
    return S_half


def Lowdin_symmetric_orthonormalization(H, S):
    """
    Lowdin's symmetric orthonormalization.
    """
    S_half = Lowdin(S)
    H_prime = S_half @ H @ S_half
    return H_prime



import numpy as np


def Jiso_to_Jtensor(Jiso):
    """Convert isotropic exchange to tensor form

    :param Jiso: scalar, isotropice exchange
    :returns:  A 3x3 matrix, the exchange paraemter in tensor form.
    """
    return np.eye(3, dtype=float) * Jiso


def Jani_to_Jtensor(Jani):
    """Convert anisotropic exchange to tensor form

    :param Jani: 3x3 matrix anisotropic exchange
    :returns:  A 3x3 matrix, the exchange paraemter in tensor form.
    """
    return np.array(Jani, dtype=float)


def DMI_to_Jtensor(D):
    """Convert DMI to tensor form

    :param D: vector, DMI.
    :returns:  A 3x3 matrix, the exchange paraemter in tensor form.
    """
    return np.array([[0, D[2], -D[1]], [-D[2], 0, D[0]], [D[1], -D[0], 0]], dtype=float)


def Jtensor_to_Jiso(Jtensor):
    """Convert tensor form to isotropic exchange

    :param Jtensor:  A 3x3 matrix, the exchange paraemter in tensor form.
    :returns: scalar, isotropice exchange
    """
    return np.average(np.diag(Jtensor))


def Jtensor_to_Jani(Jtensor):
    """Convert tensor form to anisotropic exchange

    :param Jtensor:  A 3x3 matrix, the exchange paraemter in tensor form.
    :returns: 3x3 matrix anisotropic exchange
    """
    return (Jtensor + Jtensor.T) / 2 - np.eye(3) * Jtensor_to_Jiso(Jtensor)


def Jtensor_to_DMI(Jtensor):
    """Convert tensor form to DMI

    :param Jtensor:  A 3x3 matrix, the exchange paraemter in tensor form.
    :returns: vector, DMI.
    """
    Dm = (Jtensor - Jtensor.T) / 2.0
    return np.array((Dm[1, 2], Dm[2, 0], Dm[0, 1]), dtype=float)


def decompose_J_tensor(Jtensor):
    """decompose a exchange tensor into the isotropic exchange (scalar)
    DMI (vector) and the anisotropic exchange (3x3 tensor).

    :param Jtensor:  a 3x3 matrix.
    :returns:

    """
    Jiso = np.average(np.diag(Jtensor))
    Dm = (Jtensor - Jtensor.T) / 2.0
    D = np.array((Dm[1, 2], Dm[2, 0], Dm[0, 1]), dtype=float)
    Jani = (Jtensor + Jtensor.T) / 2 - np.eye(3) * Jiso
    return Jiso, D, Jani


def combine_J_tensor(Jiso=0.0, D=np.zeros(3), Jani=np.zeros((3, 3), dtype=float)):
    """Combine isotropic exchange, DMI, and anisotropic exchange into tensor form

    :param Jiso: scalar, isotropice exchange
    :param D: vector, DMI.
    :param Jani: 3x3 matrix anisotropic exchange
    :returns:  A 3x3 matrix, the exchange paraemter in tensor form.
    """
    Jtensor = np.zeros((3, 3), dtype=complex)
    if Jiso is not None:
        Jtensor += np.eye(3, dtype=float) * Jiso
    if Jani is not None:
        Jtensor += np.array(Jani, dtype=float)
    if D is not None:
        Jtensor += np.array(
            [[0, D[2], -D[1]], [-D[2], 0, D[0]], [D[1], -D[0], 0]], dtype=float
        )
    return Jtensor


def test_J_tensor():
    Jiso = 5.0
    DMI = [0.1, 0.2, 0.3]
    Jani = [[0.1, 0.4, 0.3], [0.4, 0.1, 0.2], [0.3, 0.2, -0.2]]
    Jtensor = combine_J_tensor(Jiso, DMI, Jani)

    Jiso2, DMI2, Jani2 = decompose_J_tensor(Jtensor)
    # print(Jtensor)
    # print(f"{Jiso=},\n {DMI=}, \n{Jani=}")
    assert np.isclose(Jiso, Jiso2)
    assert np.allclose(DMI, DMI2)
    assert np.allclose(Jani, Jani2)


if __name__ == "__main__":
    test_J_tensor()



import numpy as np
from scipy.linalg import eigh


def gen_random_hermitean_matrix(n):
    A = np.random.rand(n, n) + 1j * np.random.rand(n, n)
    return A + A.conj().T


def gen_overlap_matrix(n):
    A = np.random.rand(n, n) + 1j * np.random.rand(n, n)
    return np.dot(A, A.conj().T)


def fermi_function(x, ef, beta):
    return 1.0 / (np.exp(beta * (x - ef)) + 1)


def test():
    n = 10
    A = gen_random_hermitean_matrix(n)
    S = gen_overlap_matrix(n)
    beta = 0.1
    ef = 0

    evals, evecs = eigh(A, S)

    etot = np.sum(evals * fermi_function(evals, ef, beta))

    rho = np.einsum("ib,b,jb->ij", evecs, fermi_function(evals, ef, beta), evecs.conj())

    etot2 = np.trace(np.dot(A, rho))

    print(etot, etot2)


if __name__ == "__main__":
    test()



#!/usr/bin/env python3
import copy
from ase.io import read, write
import numpy as np
from TB2J.tensor_rotate import Rxx, Rxy, Rxz, Ryx, Ryy, Ryz, Rzx, Rzy, Rzz


def rotate_atom_xyz(atoms, noncollinear=False):
    """
    given a atoms, return rotated atoms:
    atoms_1, ..., atoms_n,
    where we considered n diffeerent roation axes.

    When noncollinear == True, more rotated structures
    will be generated.
    """

    rotation_axes = [(1, 0, 0), (0, 1, 0)]
    if noncollinear:
        rotation_axes += [(1, 1, 0), (1, 0, 1), (0, 1, 1)]
    
    for axis in rotation_axes:
        rotated_atoms = copy.deepcopy(atoms)
        rotated_atoms.rotate(90, axis, rotate_cell=True)
        yield rotated_atoms


def rotate_atom_spin_one_rotation(atoms, Rotation):
    """
    roate the spin of atoms with one rotation operator
    """
    magmoms = np.array(atoms.get_initial_magnetic_moments())
    if len(magmoms.shape) == 1:
        m = np.zeros((len(magmoms), 3), dtype=float)
        m[:, 2] = magmoms
    else:
        m = magmoms
    rotated_m = Rotation.apply(m)
    atoms_copy = copy.deepcopy(atoms)
    atoms_copy.set_initial_magnetic_moments(None)
    atoms_copy.set_initial_magnetic_moments(rotated_m)
    return atoms_copy


def rotate_atom_spin(atoms):
    """
    given a atoms, return the atoms with rotated spin:
    - 'z'->'x'
    - 'z'->'y'
    - 'z'->'z'
    """
    return [rotate_atom_spin_one_rotation(atoms, R) for R in [Rzx, Rzy, Rzz]]


def rotate_atom_spin_legacy(atoms):
    """
    given a atoms, return the atoms with rotated spin:
    - 'z'->'x'
    - 'z'->'y'
    - 'z'->'z'
    """
    magmoms = np.array(atoms.get_initial_magnetic_moments())
    if len(magmoms.shape) == 1:
        m = np.zeros((len(magmoms), 3), dtype=float)
        m[:, 2] = magmoms
    else:
        m = magmoms
    atoms_copy = copy.deepcopy(atoms)
    atoms.set_initial_magnetic_moments(None)

    m_x = copy.deepcopy(m)
    m_x[:, 0] = m[:, 2]
    m_x[:, 1] = m[:, 1]
    m_x[:, 2] = m[:, 0]
    atoms_x = copy.deepcopy(atoms_copy)
    atoms_x.set_initial_magnetic_moments(m_x)

    m_y = copy.deepcopy(m)
    m_y[:, 0] = m[:, 0]
    m_y[:, 1] = m[:, 2]
    m_y[:, 2] = m[:, 1]
    atoms_y = copy.deepcopy(atoms_copy)
    atoms_y.set_initial_magnetic_moments(m_y)

    m_z = copy.deepcopy(m)
    m_z[:, 0] = m[:, 0]
    m_z[:, 1] = m[:, 1]
    m_z[:, 2] = m[:, 2]
    atoms_z = copy.deepcopy(atoms_copy)
    atoms_z.set_initial_magnetic_moments(m_z)
    return atoms_x, atoms_y, atoms_z


def check_ftype(ftype):
    if ftype in ["cif"]:
        print("=" * 40)
        print("WARNING!!!!!")
        print(
            f"{ftype} type does not contains the cell matrix explicitly. Therefore the outputted files does not give the rotation properly. Please use another format."
        )
        print("=" * 40)


def rotate_xyz(fname, ftype="xyz", noncollinear=False):
    check_ftype(ftype)
    atoms = read(fname)
    atoms.set_pbc(True)

    rotated = rotate_atom_xyz(atoms, noncollinear=noncollinear)

    for i, rotated_atoms in enumerate(rotated):
        write(f"atoms_{i+1}.{ftype}", rotated_atoms)
    write(f"atoms_0.{ftype}", atoms)

    print(f"The output has been written to the atoms_i.{ftype} files. atoms_0.{ftype} contains the reference structure.")
